/*********************************************************************
    This file is part of Siworin. Copyright © 2021 by Ruud Harmsen.

    Siworin is a simple word indexer and local search engine. See
    https://rudhar.com/sfreview/siworin/#siw04 and
    https://rudhar.com/sfreview/siworin/toolsrc/siworin .

    Siworin is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Siworin is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

/* Siworin - simple word indexer
   Written by Ruud Harmsen, July 2021

   siworin-makelst takes the output of siworin-wordsep,
   after it has been sorted to get identical words
   together. It transforms from, for example:
   word1!filenum1!offset1
   word1!filenum1!offset2
   word1!filenum2!offset3
   to:
   word1!filenum1!offset1!filenum1!offset2!filenum2!offset3

   The input is read from stdin (so it may come from a pipe),
   the output is in files, the names if which are determined
   by function siworin_fileopen_words in siworin.files.[hc] .
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "siworin.h"
#include "siworin-files.h"
#include "siworin-config.h"

static void Usage (char *progname)
{
   fprintf(stderr, "Usage:\n");
   fprintf(stderr, "%s\n", progname);
   fprintf(stderr, "%s -cCONFIGFILE\n", progname);
   fprintf(stderr, "%s -c CONFIGFILE\n", progname);
   fprintf(stderr, "%s --config CONFIGFILE\n", progname);
}

int main (int argc, char **argv)
{
   FILE *fi = stdin;
   static char line[SIWORIN_MAXWORDLEN + 24];
   static char pwrd[SIWORIN_MAXWORDLEN];
   static char word[SIWORIN_MAXWORDLEN];
   long newwordpos = 0;
   long wrdindxpos = 0;
   int fseek_done = 0;
   long occrlist_len = 0;
   long word_max_occur;
   int write_words_freq;

   if (argc > 3)
   {
      Usage(argv[0]);
      return 1;
   }
   siworin_args_to_config_path(argv);
   word_max_occur   = atoi(siworin_config_value(SIWORIN_CONFIG_INDEX_WORD_MAX_OCCUR));
   write_words_freq = atoi(siworin_config_value(SIWORIN_CONFIG_WRITE_WORDS_FREQ));

   if (siworin_fileopen_words(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "w", "w"))
      SIWORIN_FAIL("fopens");

   while (fgets(line, sizeof line, fi) != NULL)
   {
      char *seppos = strchr(line, SIWORIN_SEP);
      char *nl;
      unsigned long file_offset, byte_offset;

      /* Remove trailing newline */
      if ((nl = strrchr(line, '\n')) != 0)
         *nl = '\0';

      if (!seppos)
         seppos = line + strlen(line);

      strncpy(word, line, seppos - line);
      word[seppos - line] = '\0';
      if (strcmp(word, pwrd) != 0)
      {
         /* Close the previous line that was built up. */
         if (pwrd[0] && !fseek_done)
         {
            fprintf(siworin_fp_words, "\n");
            if (write_words_freq)
               fprintf(siworin_fp_wordfreq, "%ld\t%s\n", occrlist_len, pwrd);
         }

         newwordpos = ftell(siworin_fp_words);
         fseek_done = 0;
         occrlist_len = 1;

         sscanf(seppos,                SIWORIN_FORMAT2,      &file_offset, &byte_offset);
         fprintf(siworin_fp_words, "%s"SIWORIN_FORMAT2, word, file_offset,  byte_offset);

         wrdindxpos = ftell(siworin_fp_wordoffs);
         fprintf(siworin_fp_wordoffs, "%0*lx\n", (int)SIWORIN_OFFSET_HEXDIGITS, newwordpos);

         strcpy(pwrd, word);
      }
      else
      {
         /* Add more occurrences to the current line. */
         occrlist_len++;
         /* Write the occurrence only if there aren't too many. */
         if (occrlist_len <= word_max_occur)
         {
            /* File index/byte pos pair (referring to word start in an
               HTML file) is read, and rewritting in a shorter variable
               format, to save space in the eventually resulting file. */
            sscanf(seppos,            SIWORIN_FORMAT2, &file_offset, &byte_offset);
            fprintf(siworin_fp_words, SIWORIN_FORMAT2,  file_offset,  byte_offset);
         }
         else
         /* Only the first time per line. */
         if (!fseek_done)
         {
            /* Reset file position at that point (and truncate? not yet,
               see below), so the current line being built up, is
               effectively suppressed.
               Also suppress recording the beginning of that line in
               the index file. */
            fseek(siworin_fp_words, newwordpos, SEEK_SET);
            // ftruncate(fileno(siworin_fp_words), newwordpos);
            fseek_done = 1;
            fseek(siworin_fp_wordoffs, wrdindxpos, SEEK_SET);
            // ftruncate(fileno(siworin_fp_wordoffs), wrdindxpos);
         }
      }
   }
   /* Close also the last line of the file. */
   fprintf(siworin_fp_words, "\n");
   if (write_words_freq)
      fprintf(siworin_fp_wordfreq, "%ld\t%s\n", occrlist_len, pwrd);

   /* Very long lists of occurrences, which are probably uninteresting,
      are suppressed. But the part already written before the overflow
      is detected, may have been written beyond shorter entries written
      later on. Therefore we do an ftruncate. Only once at the end, not
      after every suppression, because ftruncate can be a rather slow
      operation.  See http://rudhar.com/sfreview/siworin/siworin11.htm
      for a longer explanation.
    */
   ftruncate(fileno(siworin_fp_words), ftell(siworin_fp_words));
   ftruncate(fileno(siworin_fp_wordoffs), ftell(siworin_fp_wordoffs));

   fclose(fi);
   siworin_fileopen_close();
   siworin_config_freeall();

   return 0;
}
